﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {

  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  using Interpreter;

  /// <summary>
  /// If the top item of the BOOLEAN stack is TRUE then this removes the second item on the EXEC stack, leaving the first
  /// item to be executed. If it is false then it removes the first item, leaving the second to be executed. This is similar to
  /// CODE.IF except that it operates on the EXEC stack. This acts as a NOOP unless there are at least two items on the EXEC stack and
  /// one item on the BOOLEAN stack.
  /// </summary>
  [PushExpression(StackTypes.Exec, "EXEC.IF", StackTypes.Boolean)]
  public class ExecIfExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      // not enough arguments on stack
      if ((interpreter.BooleanStack.Count == 0) || (interpreter.ExecStack.Count < 2)) return false;

      var condition = interpreter.BooleanStack.Pop();

      if (condition) interpreter.ExecStack.RemoveAt(interpreter.ExecStack.Count - 2);
      else interpreter.ExecStack.RemoveTop();

      return true;
    }
  }

  /// <summary>
  ///     Inserts beneath the top item of the EXEC stack a new item of the form
  ///     "( EXEC.Y <TopItem> )".
  /// </summary>
  [PushExpression(StackTypes.Exec, "EXEC.Y")]
  public class ExecYExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      // not enough arguments on stack
      if (interpreter.ExecStack.Count == 0 ||
          interpreter.Configuration.MaxPointsInProgram < 2)
        return false;

      var top = interpreter.ExecStack.Top;

      if (top is PushProgram && ((PushProgram)top).Depth == interpreter.Configuration.MaxDepth)
        return false;

      var execYExpression = ExpressionTable.GetStatelessExpression<ExecYExpression>();
      var expressions = interpreter.PoolContainer.ExpressionListPool.Get();

      expressions.Add(top);
      expressions.Add(execYExpression);

      var result = PushProgram.Create(interpreter.PoolContainer.PushProgramPool, expressions);

      interpreter.ExecStack.SetTop(result);
      interpreter.ExecStack.Add(top);

      return true;
    }
  }

  /// <summary>
  ///     Removes the second item on the EXEC stack.
  /// </summary>
  [PushExpression(StackTypes.Exec, "EXEC.K")]
  public class ExecKExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.ExecStack.Count < 2) return false;

      var top = interpreter.ExecStack.Pop();
      interpreter.ExecStack.SetTop(top);

      return true;
    }
  }

  /// <summary>
  ///     Pops 3 items from the EXEC stack, which we will call A, B, and C
  ///     (with A being the first one popped). Then pushes a list containing B and C back onto the EXEC stack, followed by
  ///     another instance of C, followed by another instance of A.
  /// </summary>
  [PushExpression(StackTypes.Exec, "EXEC.S")]
  public class ExecSExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.ExecStack.Count < 3) return false;

      // check max depth
      if ((interpreter.ExecStack.Top is PushProgram && ((PushProgram)interpreter.ExecStack.Top).Depth == interpreter.Configuration.MaxDepth) ||
          (interpreter.ExecStack.ReverseElementAt(1) is PushProgram && ((PushProgram)interpreter.ExecStack.ReverseElementAt(1)).Depth == interpreter.Configuration.MaxDepth))
        return false;

      var expression = interpreter.ExecStack.Pop(2);
      var a = expression[1];
      var b = expression[0];
      var c = interpreter.ExecStack.Top;

      var expressions = interpreter.PoolContainer.ExpressionListPool.Get();
      expressions.Add(c);
      expressions.Add(b);

      var newTop = PushProgram.Create(interpreter.PoolContainer.PushProgramPool, expressions);

      interpreter.ExecStack.SetTop(newTop);
      interpreter.ExecStack.Push(c, a);

      return true;
    }
  }

  /// <summary>
  ///     Does nothing.
  /// </summary>
  [PushExpression(StackTypes.Exec, "EXEC.NOOP")]
  public class ExecNoopExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return false;
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.WHILE", StackTypes.Boolean)]
  public class ExecWhileExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.ExecStack.IsEmpty)
        return false;

      if (interpreter.BooleanStack.IsEmpty) {
        interpreter.ExecStack.Pop();
        return true;
      }

      var booleanTop = interpreter.BooleanStack.Pop();

      if (!booleanTop) {
        interpreter.ExecStack.Pop();
        return true;
      }

      interpreter.ExecStack.Push(this, interpreter.ExecStack.Top);
      return true;
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.DO*WHILE")]
  public class ExecDoWhileExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.ExecStack.IsEmpty)
        return false;

      interpreter.ExecStack.Push(this, interpreter.ExecStack.Top);
      return true;
    }
  }

  [PushExpression(StackTypes.Exec, "EXEC.WHEN", StackTypes.Boolean)]
  public class ExecWhenExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.ExecStack.IsEmpty ||
          interpreter.BooleanStack.IsEmpty)
        return false;

      var when = interpreter.BooleanStack.Pop();

      if (!when) {
        interpreter.ExecStack.Pop();
      }

      return true;
    }
  }
}