﻿using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Generators.CodeGenerator {
  using HeuristicLab.Core;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Erc.Interfaces;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  internal static class CodeGeneratorUtils {
    internal static Expression CreateExpression(IRandom random, IReadOnlyList<string> enabledExpressions, IDictionary<string, Expression> customExpressions = null) {
      var customCount = customExpressions == null ? 0 : customExpressions.Count;
      var index = random.Next(0, enabledExpressions.Count + customCount);

      if ((index >= 0) && (index < enabledExpressions.Count)) {
        var name = enabledExpressions[index];
        return ExpressionTable.GetExpression(name);
      }

      return customExpressions.ElementAt(index - (enabledExpressions.Count - 1)).Value;
    }

    internal static Expression CreateExpressionOrErc(
      IRandom random,
      IReadOnlyList<string> enabledExpressions,
      IReadOnlyErcOptions ercOptions,
      IDictionary<string, Expression> customExpressions = null) {
      var expression = CreateExpression(random, enabledExpressions, customExpressions);
      return CreateExpressionOrErc(expression, random, ercOptions);
    }

    internal static Expression CreateExpressionOrErc(
      int index,
      IRandom random,
      IReadOnlyList<string> enabledExpressions,
      IReadOnlyErcOptions ercOptions) {
      var name = enabledExpressions[index];
      var expression = ExpressionTable.GetExpression(name);

      return CreateExpressionOrErc(expression, random, ercOptions);
    }

    private static Expression CreateExpressionOrErc(
      Expression expression,
      IRandom random,
      IReadOnlyErcOptions ercOptions) {
      var x = random.NextDouble();

      if (x < ercOptions.ErcProbability) {
        var tmp = CreateRandomErcExpression(
            ExpressionTable.TypeToAttributeTable[expression.GetType()].StackType,
            random,
            ercOptions);

        if (!(tmp is ExecNoopExpression)) {
          expression = tmp;
        }
      }

      return expression;
    }

    /// <summary>
    /// Create a ErcExpression whereby the type of the expression conforms to the passed stack type.
    /// </summary>
    /// <param name="type"></param>
    /// <param name="random"></param>
    /// <param name="pushConfiguration"></param>
    /// <returns>ErcExpression or Noop if the required ErcOptionConstants is not configured or enabled in the passed config.</returns>
    internal static Expression CreateRandomErcExpression(StackTypes type, IRandom random, IReadOnlyErcOptions ercOptions) {
      switch (type) {
        case StackTypes.Integer:
          if (ercOptions.IntegerErcOptions == null || !ercOptions.IntegerErcOptions.IsEnabled)
            return ExpressionTable.GetStatelessExpression<ExecNoopExpression>();

          var intValue = ErcUtils.GetInteger(ercOptions.IntegerErcOptions, random);

          return new IntegerPushExpression(intValue);

        case StackTypes.Float:
          if (ercOptions.FloatErcOptions == null || !ercOptions.FloatErcOptions.IsEnabled)
            return ExpressionTable.GetStatelessExpression<ExecNoopExpression>();

          var doubleValue = ErcUtils.GetDouble(ercOptions.FloatErcOptions, random);

          return new FloatPushExpression(doubleValue);

        case StackTypes.Boolean:
          return ercOptions.BooleanErcOptions == null || !ercOptions.BooleanErcOptions.IsEnabled
            ? ExpressionTable.GetStatelessExpression<ExecNoopExpression>()
            : new BooleanPushExpression(ErcUtils.GetBoolean(ercOptions.BooleanErcOptions, random));

        case StackTypes.Name:
          return ercOptions.NameErcOptions == null || !ercOptions.NameErcOptions.IsEnabled
            ? ExpressionTable.GetStatelessExpression<ExecNoopExpression>()
            : new NamePushExpression(ErcUtils.GetName(ercOptions.NameErcOptions, random));

        case StackTypes.String:
          return ercOptions.StringErcOptions == null || !ercOptions.StringErcOptions.IsEnabled
            ? ExpressionTable.GetStatelessExpression<ExecNoopExpression>()
            : new StringPushExpression(ErcUtils.GetString(ercOptions.StringErcOptions, random));

        case StackTypes.Char:
          return ercOptions.CharErcOptions == null || !ercOptions.CharErcOptions.IsEnabled
            ? ExpressionTable.GetStatelessExpression<ExecNoopExpression>()
            : new CharPushExpression(ErcUtils.GetChar(ercOptions.CharErcOptions, random));

        default:
          return ExpressionTable.GetStatelessExpression<ExecNoopExpression>();
      }
    }
  }
}
